@@ -11,22 +11,6 @@ def exists(val):
 def default(val, d):
     return val if exists(val) else d
 
-# for controlling freezing of parameters
-
-def set_module_requires_grad_(module, requires_grad):
-    for param in module.parameters():
-        param.requires_grad = requires_grad
-
-def freeze_all_layers_(module):
-    set_module_requires_grad_(module, False)
-
-def unfreeze_all_layers_(module):
-    set_module_requires_grad_(module, True)
-
-def freeze_model_and_make_eval_(model):
-    model.eval()
-    freeze_all_layers_(model)
-
 # normalization
 # they use layernorm without bias, something that pytorch does not offer
 
@@ -51,6 +35,17 @@ class Residual(nn.Module):
     def forward(self, x, *args, **kwargs):
         return self.fn(x, *args, **kwargs) + x
 
+# to latents
+
+
+class EmbedToLatents(nn.Module):
+    def __init__(self, dim, dim_latents):
+        super().__init__()
+        self.to_latents = nn.Linear(dim, dim_latents, bias=False)
+
+    def forward(self, x):
+        latents = self.to_latents(x)
+        return F.normalize(latents, dim=-1)
 
 # rotary positional embedding
 # https://arxiv.org/abs/2104.09864
@@ -298,6 +293,7 @@ class CoCa(nn.Module):
         num_tokens,
         unimodal_depth,
         multimodal_depth,
+        dim_latents = None,
         image_dim = None,
         num_img_queries=256,
         dim_head=64,
@@ -324,9 +320,6 @@ class CoCa(nn.Module):
 
         self.img_encoder = img_encoder
 
-        if exists(self.img_encoder):
-            freeze_model_and_make_eval_(self.img_encoder)
-
         # attention pooling for image tokens
 
         self.img_queries = nn.Parameter(torch.randn(num_img_queries + 1, dim)) # num image queries for multimodal, but 1 extra CLS for contrastive learning
@@ -335,6 +328,12 @@ class CoCa(nn.Module):
         self.img_attn_pool_norm = LayerNorm(dim)
         self.text_cls_norm = LayerNorm(dim)
 
+        # to latents
+
+        dim_latents = default(dim_latents, dim)
+        self.img_to_latents = EmbedToLatents(dim, dim_latents)
+        self.text_to_latents = EmbedToLatents(dim, dim_latents)
+
         # contrastive learning temperature
 
         self.temperature = nn.Parameter(torch.Tensor([1.]))
@@ -405,10 +404,7 @@ class CoCa(nn.Module):
 
         if exists(images):
             assert exists(self.img_encoder), 'img_encoder must be passed in for automatic image encoding'
-
-            self.img_encoder.eval()
-            with torch.no_grad():
-                image_tokens = self.img_encoder(images).detach()
+            image_tokens = self.img_encoder(images)
 
         # attention pool image tokens
 
@@ -458,13 +454,18 @@ class CoCa(nn.Module):
 
         # calculate caption loss (cross entropy loss)
 
-        logits = rearrange(logits, 'b c n -> b n c')
+        logits = rearrange(logits, 'b n c -> b c n')
         caption_loss = ce(logits, labels, ignore_index=self.pad_id)
         caption_loss = caption_loss * self.caption_loss_weight
 
+        # embedding to latents
+
+        text_latents = self.text_to_latents(text_embeds)
+        image_latents = self.img_to_latents(image_embeds)
+
         # calculate contrastive loss
 
-        sim = einsum('i d, j d -> i j', text_embeds, image_embeds)
+        sim = einsum('i d, j d -> i j', text_latents, image_latents)
         sim = sim * self.temperature.exp()
         contrastive_labels = torch.arange(batch, device=device)
 
